;; Internally, we always work with y coordinates as Langband does,
;; i.e. reversed w.r.t. the server's coordinates.  This limits the
;; number of places where coordinates must be transformed to inside
;; this file.

(in-package :org.langband.vanilla)

;;; !!!
(setf *print-circle* t)

(defvar *socket-input* 'not-connected)
(defvar *socket-output* 'not-connected)

(defvar *board*)
(defvar *height*)
(defvar *width*)

(defvar *verbose* nil)

(defun connect (&optional (host "localhost") (port 1111))
  (assert (eq *socket-input* 'not-connected))
  (format *debug-io* "~&host=~A port=~A~%" host port)
  (check-type host string)
  (check-type port (unsigned-byte 16))
  (setq *socket-input* (port:open-socket host port)
        *socket-output* *socket-input*))

(defun disconnect ()
  (assert (not (eq *socket-input* 'not-connected)))
  (close *socket-input*)
  (setq *socket-input* 'not-connected
        *socket-output* 'not-connected)
  t)


(defun get-token ()
  (let ((result
         (read-preserving-whitespace *socket-input*)))
    (when *verbose*
      (format *trace-output* "~&(got token: ~S)~%" result))
    result))

(defun get-line ()
  (let ((result
         (read-line *socket-input*)))
    (when *verbose*
      (format *trace-output* "~&(got line: ~S)~%" result))
    result))

(defun format-to-socket (format-control &rest format-args)
  (let ((output
         (apply #'format nil format-control format-args)))
    (when *verbose*
      ;; hack
      (assert (eql (schar output (1- (length output))) #\Newline))
      (format *trace-output* "~&(writing line: ~S)~%"
              (subseq output 0 (1- (length output)))))
    (format *socket-output* "~A" output)))

(port:defconst +whitespace+ (simple-array character (*))
  ;; Not sure about #\Newline.
  (port:mk-arr 'character
               '(#\Space #\Newline #\Tab #\Linefeed #\Return #\Page))
  "Whitespace chars.")

(port:defsubst whitespace-char-p (char)
  (find char +whitespace+ :test #'char=))

(defun skip-eol ()
  (assert (every #'whitespace-char-p (get-line))))

(defun pair (x y)
  (+ (* x 10000) y))
(defun pair-x (pair)
  (truncate pair 10000))
(defun pair-y (pair)
  (mod pair 10000))
(defun pair= (p1 p2)
  (eql p1 p2))

(defstruct sq
  pair
  char
  ;; List of pkgs that are definitely here.
  (pkgs nil)
  ;; List of pkgs that got dropped here, but have an unknown
  ;; destination.
  (pkgs-maybe nil)
  (robots '()))

(defun square-x (sq)
  (pair-x (sq-pair sq)))
(defun square-y (sq)
  (pair-y (sq-pair sq)))
(defun sq-x (sq)
  (square-x sq))
(defun sq-y (sq)
  (square-y sq))

(defun get-sq (x y)
  (aref *board* x y))

(defun read-board ()
  (setq *width* (+ (get-token) 2)
        *height* (+ (get-token) 2)
        *board* (make-array (list *width* *height*)))
  (skip-eol)
  (loop for y from (- *height* 2) downto 1
        for line = (get-line) do
        (loop for x from 1 upto (- *width* 2)
              for char = (schar line (1- x)) do
              (setf (aref *board* x y)
                    (make-sq :pair (pair x y) :char char))))
  ;; Draw borders.
  (dolist (y (list 0 (1- *height*)))
    (dotimes (x *width*)
      (setf (aref *board* x y)
            (make-sq :pair (pair x y) :char #\#))))
  (dolist (x (list 0 (1- *width*)))
    (dotimes (y *height*)
      (setf (aref *board* x y)
            (make-sq :pair (pair x y) :char #\#)))))

(defun print-board ()
  (format *debug-io* "~&")
  (loop for y from (1- *height*) downto 0 do
        (loop for x from 0 below *width*
              do (format *debug-io* "~A"
                         (sq-char (aref *board* x y))))
        (format *debug-io* "~%")))

(defvar *self*)
(defvar *maxweight*)
(defvar *weight*)
(defvar *maxmoney*)
(defvar *money*)

;; Nil generally means unknown (except for lists).
(defstruct robot
  id
  (sq nil)
  ;; Pkgs we definitely know it is carrying.
  pkgs)

(defun robot-x (r)
  (sq-x (robot-sq r)))
(defun robot-y (r)
  (sq-y (robot-sq r)))
(defun current-sq ()
  (robot-sq *self*))

(defstruct pkg
  id
  (dest nil)
  (weight nil))

(defvar *robot-hash* (make-hash-table :test 'eql))
(defvar *pkg-hash* (make-hash-table :test 'eql))

(defun get-robot (id)
  (or (gethash id *robot-hash*)
      (setf (gethash id *robot-hash*) (make-robot :id id))))
(defun get-pkg (id)
  (or (gethash id *pkg-hash*)
      (setf (gethash id *pkg-hash*) (make-pkg :id id))))
(defun destroy-pkg (pkg)
  (remhash (pkg-id pkg) *pkg-hash*)
  ;; Hm.  We might note somewhere that the id may not be used again.
  )

(defun read-robot-configuration ()
  (setq *self* (get-robot (get-token))
        *maxweight* (get-token)
        *weight* 0
        *maxmoney* (get-token)
        *money* *maxmoney*)
  (skip-eol))

(defun canonicalize-y (server-y)
  (- *height* server-y 1))
(defun decanonicalize-y (canonical-y)
  (- *height* canonical-y 1))

(defun robot-name-id (robot)
  (if (eql robot *self*)
    "you"
    (format nil "robot #~A" (robot-id robot))))
(defun pkg-name-id (pkg)
  (format nil "package #~A" (pkg-id pkg)))
(defun sq-name-coordinates (sq)
  (format nil "(~A,~A)" (sq-x sq) (decanonicalize-y (sq-y sq))))

(defun parse-action-report (string func)
  (with-input-from-string (stream string)
    (loop until (null (peek-char t stream nil nil)) do
          (assert (eql (read-char stream) #\#))
          (let ((robot (get-robot (read stream))))
            (loop with found-an-action = nil
                  until (member (peek-char t stream nil nil)
                                '(#\# nil))
                  do
                  (setq found-an-action t)
                  (let ((action-char (read-char stream)))
                    (ecase action-char
                      (#\N (funcall func robot 'Move 'N))
                      (#\S (funcall func robot 'Move 'S))
                      (#\E (funcall func robot 'Move 'E))
                      (#\W (funcall func robot 'Move 'W))
                      (#\P (funcall func robot
                                    'Pick (get-pkg (read stream))))
                      (#\D (funcall func robot
                                    'Drop (get-pkg (read stream))))
                      (#\X
                       (let ((xpos (read stream))
                             (ypos (canonicalize-y
                                    (progn (peek-char t stream)
                                           (assert
                                            (eql (read-char stream) #\Y))
                                           (read stream)))))
                         (funcall func robot
                                  'Appear (get-sq xpos ypos)))))

		    (bit-flag-add! *redraw* +print-basic+))
                  finally (unless found-an-action
                            (funcall func robot 'Nop nil))
                  )))))

(defun get-neighbor (sq dir)
  (multiple-value-bind (dx dy) (ecase dir
                                 (N (values 0 -1))
                                 (E (values 1 0))
                                 (S (values 0 1))
                                 (W (values -1 0)))
    (let* ((pair (sq-pair sq))
           (x (pair-x pair))
           (y (pair-y pair)))
      (get-sq (+ x dx) (+ y dy)))))

(defun format-message! (format-control &rest format-args)
  (let ((text (apply #'format nil format-control format-args)))
    (print-message! text)
    (when *verbose*
      (format *debug-io* "~&>>> ~A~%" text))))

(defun %update-spot (dungeon square)
  (let* ((x (sq-x square))
	 (y (sq-y square))
	 (mon (cave-monsters dungeon x y)))

    (dolist (i mon)
      (setf (amon.seen-by-player? i) t))

    (note-spot! dungeon x y)
    (light-spot! dungeon x y)
    ;; bad
    (c-term-fresh!)))

;; FIXME: reselect-floor-item and reselect-monster might cons *a lot*.
;; It would probably be useful to have some pool of recyclable
;; objects.

(defun reselect-floor-item (dungeon sq)
  (setf (cave-objects dungeon (sq-x sq) (sq-y sq)) nil)
  ;; FIXME: There should be some marker when someone has dropped a
  ;; package with unknown destination, too.
  (when (sq-pkgs sq)
    (let* ((obj (get-object-kind *variant* "package"))
           (aobj (create-aobj-from-kind obj :variant *variant* :amount 1)))
      (let-floor-carry! dungeon (sq-x sq) (sq-y sq) aobj)))
  (%update-spot dungeon sq))

(defun reselect-monster (dungeon player sq)
  ;; FIXME: Does this remove the player, too?  If yes, is that bad?
  (setf (cave-monsters dungeon (sq-x sq) (sq-y sq)) nil)
  
  (cond ((pair= sq (current-sq))
         ;;(warn "Square ~a,~a is occupied by player." (sq-x sq) (sq-y sq))
         (setf (location-x player) (sq-x sq)
               (location-y player) (sq-y sq))
         ;;(bit-flag-add! *update* +pl-upd-update-view+ +pl-upd-distance+
         ;;               +pl-upd-panel+ +pl-upd-update-flow+)
	 )
        ((sq-robots sq)
         ;;(warn "Square ~a,~a is occupied by monsters." (sq-x sq) (sq-y sq))
         (let* ((mon (get-monster-kind *variant* "robot"))
                (amon (produce-active-monster *variant* mon)))
           (setf (amon.seen-by-player? amon) t)
           (place-single-monster! dungeon player amon
                                  (sq-x sq) (sq-y sq) nil)))
        (t ;;(warn "Square ~a,~a is empty." (sq-x sq) (sq-y sq))
         ))
  (%update-spot dungeon sq))

(defun move-robot (dungeon player robot to)
  (let ((from (robot-sq robot)))
    (when from
      (assert (member robot (sq-robots from)))
      (setf (sq-robots from) (remove robot (sq-robots from))))
    (setf (robot-sq robot) to)
    (when to
      (assert (not (member robot (sq-robots to))))
      (push robot (sq-robots to)))
    (when from (reselect-monster dungeon player from))
    (when to (reselect-monster dungeon player to))))

(defun destroy-robot (dungeon player robot)
  (move-robot dungeon player robot nil)
  (remhash (robot-id robot) *robot-hash*)
  ;; Hm.  We might note somewhere that the id may not be used again.
  )

(defun extract-initial-player-positions (dungeon player string)
  (parse-action-report
   string
   (lambda (robot action arg)
     (ecase action
       (Nop t)
       (Move t)
       (Pick t)
       (Drop t)
       (Appear
        (cond ((robot-sq robot)
;               (warn
;                "This shouldn't happen: Two 'appear's for the same robot.")
               (move-robot dungeon player robot arg))
              ((eql robot *self*)
               (format-message! "You enter the game at ~A."
                                (sq-name-coordinates arg))
               (place-player! dungeon player (sq-x arg) (sq-y arg))
               (move-robot dungeon player robot arg))
              (t
               (move-robot dungeon player robot arg))))))))

(defvar *current-round-robots* '())

(defun process-action-report (dungeon player string)
  (let ((next-round-robots '()))
    (parse-action-report
     string
     (lambda (robot action arg)
       (pushnew robot next-round-robots)
       (ecase action
         (Nop t)
         (Move
          (let ((from (robot-sq robot))
                (to (get-neighbor (robot-sq robot) arg)))
;            (warn "Got move message from (~S,~S) to (~S,~S)."
;                  (sq-x from) (sq-y from)
;                  (sq-x to) (sq-y to))
            (move-robot dungeon player robot to)
            ;; This seems to be tricky.  The exact same calls are
            ;; already being made during move-robot above, but these
            ;; are required nonetheless.
            (%update-spot dungeon from)
            (%update-spot dungeon to)
	    ))
         (Pick
          (assert (not (member arg (robot-pkgs robot))))
          (let ((sq (robot-sq robot)))
            (setf (sq-pkgs sq) (remove arg (sq-pkgs sq))
                  (sq-pkgs-maybe sq) (remove arg (sq-pkgs-maybe sq)))
            (if (eql robot *self*)
              (progn
                (assert (pkg-weight arg))
                (incf *weight* (pkg-weight arg))
                (format-message! "You pick up ~A." (pkg-name-id arg)))
              (format-message! "~@(~A~) picks up ~A."
                               (robot-name-id robot)
                               (pkg-name-id arg)))
            (push arg (robot-pkgs robot))
            (reselect-floor-item dungeon sq)))
         (Drop
          (setf (robot-pkgs robot) (remove arg (robot-pkgs robot)))
          (let ((sq (robot-sq robot)))
            (cond ((null (pkg-dest arg))
                   (assert (not (eql robot *self*)))
                   (format-message! "~@(~A~) drops ~A, possibly delivering it."
                                    (robot-name-id robot) (pkg-name-id arg))
                   (push arg (sq-pkgs-maybe sq))
                   (reselect-floor-item dungeon sq))
                  ((pair= sq (pkg-dest arg))
                   (if (eql robot *self*)
                     (format-message! "You deliver ~A." (pkg-name-id arg))
                     (format-message! "~@(~A~) delivers ~A."
                                      (robot-name-id robot) (pkg-name-id arg)))
                   (destroy-pkg arg))
                  (t
                   (if (eql robot *self*)
                     (format-message! "You drop ~A." (pkg-name-id arg))
                     (format-message! "~@(~A~) drops ~A."
                                      (robot-name-id robot)
                                      (pkg-name-id arg)))
                   (push arg (sq-pkgs sq))
                   (reselect-floor-item dungeon sq)))
            (when (eql robot *self*)
              (assert (pkg-weight arg))
              (decf *weight* (pkg-weight arg)))))
         (Appear
          (cond ((robot-sq robot)
                 ;; This happens for all players who are already in
                 ;; the game when we join.
                 (unless (eql robot *self*)
                   (format-message! "~@(~A~) is at ~A."
                                    (robot-name-id robot)
                                    (sq-name-coordinates arg)))
                 (move-robot dungeon player robot arg))
                ((eql robot *self*)
                 (place-player! dungeon player (sq-x arg) (sq-y arg))
                 (unless (null (sq-robots arg))
                   (format-message! "You are on top of ~A."
                                    (robot-name-id (first (sq-robots arg)))))
                 (move-robot dungeon player robot arg))
                (t
                 (if (null (sq-robots arg))
                   (format-message! "~@(~A~) has entered the game at ~A."
                                    (robot-name-id robot)
                                    (sq-name-coordinates arg))
                   (format-message!
                    "~@(~A~) has entered the game on top of ~A."
                    (robot-name-id robot)
                    (robot-name-id (first (sq-robots arg)))))
                 (move-robot dungeon player robot arg)))))))

    (bit-flag-add! *update* +pl-upd-panel+)
    (when-bind (deaths (set-difference *current-round-robots*
                                       next-round-robots))
               (dolist (robot deaths)
                 (format-message! "~@(~A~) died." (robot-name-id robot))
                 (destroy-robot dungeon player robot)))
    (setq *current-round-robots* next-round-robots)
    ;;(format *debug-io* "~&I am now ~S~%" *self*)
    ))

(defvar *rounds-completed*)

(defun process-pkg-report (dungeon string)
  (let ((was-maybe-square (and (sq-pkgs-maybe (current-sq))
                               (not (sq-pkgs (current-sq))))))
    (with-input-from-string (stream string)
      (setf (sq-pkgs (current-sq)) nil
            (sq-pkgs-maybe (current-sq)) nil)
      (loop for pkg-id = (read stream nil nil)
            while pkg-id
            for pkg = (get-pkg pkg-id) do
            (setf (pkg-dest pkg) (get-sq (read stream)
                                         (canonicalize-y (read stream)))
                  (pkg-weight pkg) (read stream))
            (push pkg (sq-pkgs (current-sq)))))
    (reselect-floor-item dungeon (current-sq))
;    (let ((*print-level* 3))
;      (warn "Package report: ~S" (sq-pkgs (current-sq))))
    (let ((pkgs (sq-pkgs (current-sq))))
      (cond ((and was-maybe-square (null pkgs))
             ;; This does not make sense as we don't have a "maybe"-marker.
             ;;(format-message! "There are no packages here.")
             )
            ((cdr pkgs)
             (format-message! "There are some packages here."))
            (pkgs
             (format-message! "There is a package here."))))))


;; Dungeon object not available yet, as its dimensions cannot be known
;; before joining.
(defun join-game ()
  (format-to-socket "Player~%")
  (read-board)
  (read-robot-configuration))

(defun send-move-to-server (dungeon player bid move arg)
  (decf *money* bid)
  (ecase move
    (Move (format-to-socket "~A Move ~A~%"
                            bid (ecase arg
                                  (N #\N)
                                  (E #\E)
                                  (S #\S)
                                  (W #\W))))
    (Wait (assert (null arg))
          (format-to-socket "~A Drop~%" bid))
    ((Pick Drop)
     (assert (not (endp arg)))
     (format-to-socket "~A ~A ~{~A~^ ~}~%"
                       bid (ecase move
                             (Pick "Pick")
                             (Drop "Drop"))
                       arg)))
  (handler-case
      (let ((action-report (get-line)))
        (if (string= action-report
                       (format nil "Robot ~A died." (robot-id *self*)))
          (setf (player.dead-p player) t
                (player.dead-from player) "death notice from server"
                (player.leaving-p player) :dead)
          (progn
            (process-action-report dungeon player action-report)
            (process-pkg-report dungeon (get-line))
            (incf *rounds-completed*))))
    (end-of-file (cond)
      (declare (ignore cond))
      (setf (player.dead-p player) t
            (player.dead-from player) "EOF from server"
            (player.leaving-p player) :dead))))

#|
(defun play-continuously (max-rounds &key (quiet nil))
  (let ((end-after (if (eq max-rounds t)
                     0
                     (+ *rounds-completed* max-rounds))))
    (unless quiet
      (format *debug-io* "~&Round: "))
    (handler-case (loop while (or (eq max-rounds t)
                                  (< *rounds-completed* end-after))
                        do
                        (unless quiet
                          (format *debug-io* "~A " (1+ *rounds-completed*)))
                        (play-round))
      (end-of-file (condition)
        (declare (ignore condition))
        (format *debug-io* "~&Encountered EOF in round ~A, exiting.~%"
                (1+ *rounds-completed*))
        (disconnect)))
    *rounds-completed*))
|#
